/*
 * shoehorn.c --	Host-side bootstrap loader for Armadillo-9 boards.
 * 
 * Copyright (C) 2005 Sunbit System Co.,Ltd. (ported NetBSD/i386)
 * Copyright (C) 2001 Atmark Techno, Inc.
 * Copyright (c) 2000 Blue Mug, Inc.  All Rights Reserved.
 * Copyright (c) 1999 Ben Williamson <benw@pobox.com>
 *
 * This program loads a piece of code (see loader.c) into the target's
 * internal SRAM.  The loader then does our bidding to initialise and
 * detect hardware, load a kernel, ramdisk image, and boot.
 *
 * To add support for a new board type, add a new command-line option
 * that sets the hardware variable, and arrange for it to call
 * init_myboardname() instead of init_anvil().
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>
#include <errno.h>

#include "ioregs.h"
#include "serial.h"
#include "util.h"

#ifdef MSW
void messagebox (const char *str);
void consoleout (const char *str);
#endif

#define _POSIX_SOURCE	1

#define MAKE_VERSION(ver) "shoehorn v" #ver " built " \
	__TIME__", "__DATE__ \
	" (CVS: $Id: shoehorn.c,v 1.10 2001/01/17 22:50:56 miket Exp $)";
#if	defined(__NetBSD__)
static const char version[] 	= MAKE_VERSION(3.4-armadillo-9-netbsdbase);
#else
static const char version[] 	= MAKE_VERSION(3.4-armadillo-9);
#endif


#define SRAM_SIZE	0x800		/* 2kB of SRAM for loader.bin */
#define START_CHAR	'<'
#define END_CHAR	'>'
#define MAX_FRAGS	100
#define DRAM_START	0xc0000000

#define MINORBITS	8
#define MKDEV(ma,mi)	(((ma) << MINORBITS) | (mi))
#define RAMDISK_MAJOR	1

/* architecture numbers are defined by the ARM kernel */
//#define ARCH_NUMBER_EDB7211	50
#define ARCH_NUMBER_ARMADILLO9	130

#define PARAM_OFFSET	0x00020000	/* immediately after video RAM */
  #define PARAM_U1_OFFSET	0x00020000
  #define PARAM_U2_OFFSET	0x00020100
  #define PARAM_CMDLINE_OFFSET	0x00020500
  #define PARAM_END		0x00020900
  #define PARAM_SIZE		(256 + 1024 + 1024)

#define KERNEL_OFFSET	0x00218000	/* beginning of kernel image */
#define PAGE		0x1000

#define INITRD_START	0xc0400000

#ifndef MSW
#define ETH_STEP	1024
#endif

static int hardware = 0;
static int terminal = 0;
static int boot     = 0;

struct option options[] = {
	{ "armadillo9",	0, &hardware,	'e' },
	{ "boot",	0, &boot,	1 },
	{ "initrd",	1, 0,		'i' },
	{ "kernel",	1, 0,		'k' },
	{ "loader",	1, 0,		'l' },
#ifndef MSW
	{ "netif",	1, 0,		'n' },
#endif
	{ "port",	1, 0,		'p' },
	{ "terminal",	0, &terminal,	1 },
	{ "version",	0, 0,		'v' },
	{ 0,		0, 0,		0 }
};

/* this magic is documented in the GNU cpp info files; we use
   it here to let a Makefile or build script specify the path
   to loader.bin. */
#ifdef LOADERPATH
#define loaderpath(s) str(s)
#else
#define loaderpath(s)
#endif
#define str(s) #s

static int arch_number	= -1;
static char *initrd	= "initrd";
static char *kernel	= "Image";
static char *loader	= loaderpath(LOADERPATH) "loader.bin";
#ifndef MSW
static char *netif	= "eth0";
#endif
#if	defined(__NetBSD__)
static char *port	= "/dev/tty00";
#else
static char *port	= "/dev/ttyS0";
#endif

char *progname		= "UNKNOWN";

char kargs[256];
unsigned char remotemac[6];

struct fragment {
	unsigned int	start;
	unsigned int	size;
} frag_list[MAX_FRAGS], *frag;

/*
 * Print usage instructions and exit
 */
void
usage_and_exit(void)
{
#ifndef MSW
	printf("Usage: %s [options] [kernel command line]\n"
	       "Available options (defaults):\n"
	       "        --boot\n"
	       "        --armadillo9\n"
	       "        --initrd (%s)\n"
	       "        --kernel (%s)\n"
	       "        --loader (%s)\n"
	       "        --netif (%s)\n"
	       "        --port (%s)\n"
	       "        --terminal\n"
	       "        --version\n",
	       progname, initrd, kernel, loader, netif, port);
#endif
	exit(1);
}

/*
 * Parse the command line options
 */
static void
parse_command_line(int argc, char **argv)
{
	int c;
	
	while (1) {
#if	defined(__NetBSD__)
		c = getopt_long(argc, argv, "iklnp", options, NULL);
#else
		c = getopt_long_only(argc, argv, "iklnp", options, NULL);
#endif
		if (c == -1) {
			break;
		}
		switch (c) {
		case 0:
			break;
		case 'i':
			initrd = optarg;
			break;
		case 'k':
			kernel = optarg;
			break;
		case 'l':
			loader = optarg;
			break;
#ifndef MSW
		case 'n':
			netif = optarg;
			break;
#endif
		case 'p':
			port = optarg;
			break;
		case 'v':
#ifndef MSW
			puts(version);
#else
			messagebox(version);
#endif
			exit(0);
		default:
			usage_and_exit();
		}
	}
	if (hardware == 0) {
#ifndef MSW
		fprintf(stderr, "Need to specify hardware type (hint: --armadiloo)\n");
#else
		messagebox("Need to specify hardware type (hint: --armadiloo)\n");
#endif
		usage_and_exit();
	}
}

/*
 * concatenating what's left into kargs, but after root privs are dropped
 */
static void
build_kargs(int argc, char **argv)
{
	int remspace;
	int prependspace = 0;	/* 0 on first arg, else 1 */

	memset(kargs, 0, sizeof kargs);
	while (optind < argc) {
		remspace = sizeof(kargs) - strlen(kargs) - 1;
		/* add if it fits, exit if too large */
		if (remspace >= strlen(argv[optind]) + prependspace) {
			/* will fit */
			if (prependspace)	strcat(kargs, " ");
			strcat(kargs, argv[optind++]);
		} else {
#ifndef MSW
			printf("Kernel options too long\n");
#else
			messagebox("Kernel options too long\n");
#endif
			exit(1);
		}
		prependspace = 1;
	}
	kargs[sizeof(kargs) - 1] = '\0';
}

void target_write(unsigned addr, const char *buf,
		  unsigned size, unsigned progress)
{
	/* XXX this is kind of nasty */
		target_write_block(addr, buf, size, progress);
}

/*
 * Write an object to the target, spanning multiple DRAM fragments
 * Returns final address
 */
unsigned int
target_write_fragmented(unsigned int addr, char *buf, unsigned int size)
{
	struct fragment *frag;
	unsigned int frag_end = 0;
	int progress = 0;
	
	for (frag = &frag_list[1]; frag->size != 0; frag++) {
		frag_end = frag->start + frag->size;
		if ((frag->start <= addr) && (addr < frag_end)) {
			break;
		}
	}
	while (size > 0) {
		int step = min(size, frag_end - addr);
		if (step <= 0) {
#ifndef MSW
			printf("Insufficient DRAM space\n");
#else
			messagebox("Insufficient DRAM space\n");
#endif
			exit(1);
		}
		target_write(addr, buf, step, progress);
		addr += step;
		buf += step;
		progress += step;
		size -= step;
		if (size > 0) {
			frag++;
			addr = frag->start;
			frag_end = frag->start + frag->size;
		}
	}
	if (addr == frag_end) {
		addr = frag[1].start;
	}
	return addr;
}


void
ping(void)
{
	char c;

#ifndef MSW
	printf("Pinging loader\n");
#else
	char str[256];
	consoleout("Pinging loader\n");
#endif
	put_char('a');
	c = get_char();
	if (c != '+') {
#ifndef MSW
		printf("Expected acknowledge '+' but got '%c'\n", c);
#else
		sprintf(str, "Expected acknowledge '+' but got '%c'\n", c);
		messagebox (str);
#endif
		exit(1);
	}
}

/*
 * This definition of param_struct is copied from include/asm-arm/setup.h
 * in Linux 2.4.0-test6-rmk5.  Changes made after that kernel should
 * hopefully remain backwards compatible.
 *
 * Note that these will have to be changed to e.g. 'u32' to work on a 64-
 * bit development workstation.
 */
#define COMMAND_LINE_SIZE 1024

struct param_struct {
    union {
	struct {
	    unsigned long page_size;		/*  0 */
	    unsigned long nr_pages;		/*  4 */
	    unsigned long ramdisk_size;		/*  8 */
	    unsigned long flags;		/* 12 */
#define FLAG_READONLY	1
#define FLAG_RDLOAD	4
#define FLAG_RDPROMPT	8
	    unsigned long rootdev;		/* 16 */
	    unsigned long video_num_cols;	/* 20 */
	    unsigned long video_num_rows;	/* 24 */
	    unsigned long video_x;		/* 28 */
	    unsigned long video_y;		/* 32 */
	    unsigned long memc_control_reg;	/* 36 */
	    unsigned char sounddefault;		/* 40 */
	    unsigned char adfsdrives;		/* 41 */
	    unsigned char bytes_per_char_h;	/* 42 */
	    unsigned char bytes_per_char_v;	/* 43 */
	    unsigned long pages_in_bank[4];	/* 44 */
	    unsigned long pages_in_vram;	/* 60 */
	    unsigned long initrd_start;		/* 64 */
	    unsigned long initrd_size;		/* 68 */
	    unsigned long rd_start;		/* 72 */
	    unsigned long system_rev;		/* 76 */
	    unsigned long system_serial_low;	/* 80 */
	    unsigned long system_serial_high;	/* 84 */
	    unsigned long mem_fclk_21285;       /* 88 */ 
	} s;
	char unused[256];
    } u1;
    union {
	char paths[8][128];
	struct {
	    unsigned long magic;
	    char n[1024 - sizeof(unsigned long)];
	} s;
    } u2;
    char commandline[COMMAND_LINE_SIZE];
};


void
target_write_params(unsigned long initrd_start, unsigned long initrd_size)
{
	struct param_struct ps;
#ifdef MSW
	char str[256];
#endif

	/* sanity checking to keep updates from breaking anything */
	assert(PARAM_SIZE == PARAM_END - PARAM_OFFSET);
	assert(sizeof (struct param_struct) <= PARAM_SIZE);

	/* XXX these need to be generalized for different HW setups */
	/* XXX they are also endian specific */
	memset(&ps, 0, sizeof ps);
	/* printf("- page_size: %d\n", PAGE); */
	ps.u1.s.page_size = PAGE;
#ifndef MSW
	printf("- nr_pages (all banks): %d\n", 4096);
#else
	sprintf(str, "- nr_pages (all banks): %d\n", 4096);
	consoleout (str);
#endif
	ps.u1.s.nr_pages = 4096;
	/* XXX */
	ps.u1.s.flags = FLAG_READONLY | FLAG_RDLOAD | FLAG_RDPROMPT;
#ifndef MSW
	printf("- rootdev: (RAMDISK_MAJOR, 0)\n");
#else
	consoleout("- rootdev: (RAMDISK_MAJOR, 0)\n");
#endif
	ps.u1.s.rootdev = MKDEV(RAMDISK_MAJOR, 0);
#ifndef MSW
	printf("- pages_in_bank[0]: %d\n", 2048);
#else
	sprintf(str, "- pages_in_bank[0]: %d\n", 2048);
	consoleout (str);
#endif
	ps.u1.s.pages_in_bank[0] = 2048;
#ifndef MSW
	printf("- pages_in_bank[1]: %d\n", 2048);
#else
	sprintf(str, "- pages_in_bank[1]: %d\n", 2048);
	consoleout (str);
#endif
	ps.u1.s.pages_in_bank[1] = 2048;
#ifndef MSW
	printf("- initrd_start: 0x%lx\n", initrd_start);
#else
	sprintf(str, "- initrd_start: 0x%lx\n", initrd_start);
	consoleout (str);
#endif
	ps.u1.s.initrd_start = initrd_start;
#ifndef MSW
	printf("- initrd_size: 0x%lx\n", initrd_size);
#else
	sprintf(str, "- initrd_size: 0x%lx\n", initrd_size);
	consoleout (str);
#endif
	ps.u1.s.initrd_size = initrd_size;
#ifndef MSW
	printf("- ramdisk_size: 0x%lx\n", initrd_size);
#else
	sprintf(str, "- ramdisk_size: 0x%lx\n", initrd_size);
	consoleout (str);
#endif
	ps.u1.s.ramdisk_size = initrd_size;

	/* XXX these writes overlap */
	print_size(DRAM_START + PARAM_OFFSET, PARAM_SIZE);
	target_write_fragmented(DRAM_START + PARAM_OFFSET,
		(char*) &ps, PARAM_SIZE);
	target_write_fragmented(DRAM_START + PARAM_CMDLINE_OFFSET,
		kargs, sizeof(kargs));
}

	
void
init_armadillo9(void)
{
#ifndef MSW
	printf("- flushing cache/TLB\n");
#else
	consoleout("- flushing cache/TLB\n");
#endif
	put_char('4');

	printf("- Setting up GPIO\n");
	target_write_word(GPIO_PADR, 0x08);
	target_write_word(GPIO_PADDR, 0x08);
	target_write_word(GPIO_PBDDR, 0x20);
	target_write_word(GPIO_PCDDR, 0x00);
	target_write_word(GPIO_PDDDR, 0x00);
	target_write_word(GPIO_PEDR, 0x00);
	target_write_word(GPIO_PEDDR, 0xff);
	target_write_word(GPIO_PFDDR, 0x00);
	target_write_word(GPIO_PGDDR, 0x0c);
	target_write_word(GPIO_PHDDR, 0x00);

	printf("- Setting up flash at CS6\n");
	target_write_word(SMCBCR6, 0x18004d23);
	printf("- Setting up Miscellaneous(8bit-width) at CS1\n");
	target_write_word(SMCBCR1, 0x000047e3);
	printf("- Setting up Miscellaneous(16bit-width) at CS2\n");
	target_write_word(SMCBCR2, 0x100047e3);

	printf("- 200MHz core clock\n");
	put_char('C');
	sleep(1); // Special wait
	usleep(100000);

	printf("- Setting up SDRAM at SDCS0\n");
	target_write_word(SDRAMCFG0, 0x210028);
	target_write_word(SDRAMGLOBALCFG, 0x80000003);
	usleep(10000);
	target_write_word(SDRAMGLOBALCFG, 0x80000001);
	target_write_word(SDRAMREFRESHTIME, 10);
	usleep(10000);
	target_write_word(SDRAMREFRESHTIME, 6400);
	target_write_word(SDRAMGLOBALCFG, 0x80000002);
	target_read_word(0xc0008800);
	target_read_word(0xc1008800);
	target_read_word(0xc4008800);
	target_read_word(0xc5008800);
	target_read_word(0xc8008800);
	target_read_word(0xc9008800);
	target_read_word(0xcc008800);
	target_read_word(0xcd008800);
	target_write_word(SDRAMGLOBALCFG, 0x80000000);

#ifndef MSW
	printf("- enabling UART2\n");
#else
	consoleout("- enabling UART2\n");
#endif
	target_write_word(SYSCON_SWLOCK, 0xaa);
	target_write_word(SYSCON_DEVCFG,
			  target_read_word(SYSCON_DEVCFG) | 0x100000);

	printf("- Turn on GREEN LED\n");
	target_write_word(GPIO_PEDR, 0x1);

#ifndef MSW
	printf("Switching to 115200 baud\n");
#else
	consoleout("Switching to 115200 baud\n");
#endif
	put_char('B');

#ifndef MSW
	serial_baud(B115200);
#else
	serial_baud(115200);
#endif
}

void
post_armadillo9(void)
{
#ifndef MSW
	printf("Setting up UART2\n");
#else
	consoleout("Setting up UART2\n");
#endif
	target_write_word(UART2CR, 0);
	target_write_word(UART2RSR, 0);
	target_write_word(UART2CR_M, 0);
	target_write_word(UART2CR_L, 95);
	target_write_word(UART2CR_H, 0x60);
	target_write_word(UART2CR, 1);
}


void
detect_dram(void)
{
	unsigned int start, size, total_size;
#ifdef MSW
	char str[256];
#endif

	put_char('d');
#ifndef MSW
	printf("- %d bits wide\n", get_char());
#else
	sprintf(str, "- %d bits wide\n", get_char());
	consoleout (str);
#endif
	frag_list[0].start = 0;	/* item 0 is a dummy, list starts at 1 */
	frag_list[0].size = 0;
	total_size = 0;
	frag = frag_list;
	size = get_word();
	while (1) {
		start = get_word();
		if (start == 0) {
			break;
		}
		if (start == frag->start + frag->size) {
			frag->size += size;
		} else {
			frag++;
			if (frag >= &frag_list[MAX_FRAGS-1]) {
#ifndef MSW
				printf("Too many DRAM fragments\n");
#else
				messagebox("Too many DRAM fragments\n");
#endif
				exit(1);
			}
			frag->start = start;
			frag->size = size;
		}
		total_size += size;
	}
	frag++;
	frag->start = frag->size = 0;
	for (frag = &frag_list[1]; frag->size != 0; frag++)
		print_size(frag->start, frag->size);
#ifndef MSW
	printf("Total DRAM: %dkB\n", total_size / 1024);
#else
	sprintf(str, "Total DRAM: %dkB\n", total_size / 1024);
	consoleout (str);
#endif
}


void perror_usage_exit(const char *s)
{
#ifndef MSW
	fprintf(stderr, "%s: ", progname);
#else
	char str[256];
	sprintf(str, "%s: ", progname);
#endif
	if (errno > sys_nerr - 1) {
#ifndef MSW
		fprintf(stderr, "%s: unknown error # %d\n", s, errno);
#else
		sprintf(str + strlen(str), "%s: unknown error # %d\n", s, errno);
#endif
	} else {
#ifndef MSW
		fprintf(stderr, "%s: %s\n", s, sys_errlist[errno]);
#else
		sprintf(str + strlen(str), "%s: %s\n", s, sys_errlist[errno]);
#endif
	}
#ifdef MSW
	messagebox (str);
#endif
	usage_and_exit();
}


/*
 * Send loader.bin to the target
 * Initialise some target registers, for Anvil hardware
 * Detect what DRAM is present, building a list of fragments
 * Load the kernel image and initrd into available fragments
 * Write a kernel parameter block with command-line arguments
 * Jump to the kernel entry point
 */
#ifndef MSW
int
main(int argc, char **argv)
#else
int
shmain(int argc, char **argv)
#endif
{
#ifndef MSW
	unsigned char *initrd_buf, *kernel_buf, *loader_buf;
#else
	char str[256];
	unsigned char *kernel_buf, *loader_buf;
#endif
	unsigned kernel_size, initrd_size, loader_size;
	unsigned long kernel_end, initrd_start = INITRD_START, size;
#if	!defined(MSW) && !defined(__NetBSD__)
	uid_t ruid, euid, suid;
	int getresuid(uid_t *, uid_t *, uid_t *);	/* linux only????? */
#endif
	
	progname = strrchr(argv[0], '/');
	if (progname == NULL) {
		progname = argv[0];
	} else {
		progname++;
	}
	parse_command_line(argc, argv);

	/* make output to stdout immediately visible */
	setbuf(stdout, NULL);

#ifndef MSW
	/* drop root privs if setuid root */
	if ( (geteuid() == 0) && (getuid() != 0) ) {
		printf("Dropping root privileges\n");
		if (setuid(getuid()) != 0) {
			fprintf(stderr, "%s: error in droping root-privs\n",
								progname);
			exit(EXIT_FAILURE);
		}
		/* FIXME: make it portable to systems w/o saved uid */
		/* check if root privs are really dropped */
#if	!defined(__NetBSD__)
		if (getresuid(&ruid, &euid, &suid) != 0) {
			fprintf(stderr, "%s: could not check if root-privs "
						"were dropped\n", progname);
			exit(EXIT_FAILURE);
		}
		if ( (ruid == 0) || (euid == 0) || (suid == 0) ) {
			fprintf(stderr, "%s: not all root-privs were dropped\n"
					"\truid: %d, euid: %d, suid: %d\n",
				       	progname, ruid, euid, suid);
			exit(EXIT_FAILURE);
		}
#endif
	}
#endif

	/* fill in kargs *after* dropping privileges */
	build_kargs(argc, argv);

	/* slurp files into buffers */
	loader_size = SRAM_SIZE;  /* must allocate at least SRAM_SIZE bytes */
	read_file(loader, &loader_buf, &loader_size);
	kernel_size = 0;
	read_file(kernel, &kernel_buf, &kernel_size);
	initrd_size = 0;
#ifndef MSW
	read_file(initrd, &initrd_buf, &initrd_size);
#endif

	/* make sure loader isn't too big */
	if (loader_size > SRAM_SIZE) {
#ifndef MSW
		fprintf(stderr, "%s: loader too large (limit %d bytes)\n",
			progname, SRAM_SIZE);
#else
		sprintf(str, "%s: loader too large (limit %d bytes)\n",
			progname, SRAM_SIZE);
		messagebox (str);
#endif
		exit(1);
	}
	if (loader_size > SRAM_SIZE - 0x100)
#ifndef MSW
		fprintf(stderr, "%s: warning: loader stack might clobber code\n",
				progname);
#else
		{
			sprintf(str, "%s: warning: loader stack might clobber code\n",
					progname);
			consoleout (str);
		}
#endif

	/* open serial port and start talking to hardware */
	serial_open(port);
#ifndef MSW
	printf("Waiting for target - press Wakeup now.\n");
#else
	consoleout("Waiting for target - press Wakeup now.\n");
#endif
	while (get_char() != START_CHAR)
	  ;
#ifndef MSW
	printf("Writing SRAM loader...\n");
#else
	consoleout("Writing SRAM loader...\n");
#endif
	put_block(loader_buf, SRAM_SIZE);
	if (get_char() != END_CHAR) {
#ifndef MSW
		printf("Expected end character '%c'\n", END_CHAR);
#else
		sprintf(str, "Expected end character '%c'\n", END_CHAR);
		messagebox (str);
#endif
		exit(1);
	}
	ping();
	
	switch(hardware) {
	case 'e':
		arch_number = ARCH_NUMBER_ARMADILLO9;
#ifndef MSW
		printf("Initialising Armadillo-9 hardware:\n");
#else
		consoleout("Initialising Armadillo-9 hardware:\n");
#endif
		init_armadillo9();
		break;
	default:
#ifndef MSW
		printf("Internal error - invalid hardware value\n");
#else
		messagebox("Internal error - invalid hardware value\n");
#endif
		exit(1);
	}
	assert(arch_number > 0);
	ping();

#ifndef MSW
	printf("Detecting DRAM\n");
#else
	consoleout("Detecting DRAM\n");
#endif
	detect_dram();
	
#ifndef MSW
	printf("Loading %s:\n", kernel);
#else
	sprintf(str, "Loading %s:\n", kernel);
	consoleout (str);
#endif
	print_size(DRAM_START + KERNEL_OFFSET, kernel_size);
	kernel_end = target_write_fragmented(DRAM_START + KERNEL_OFFSET,
		kernel_buf, kernel_size);
	free(kernel_buf);

	/* Find a start address for initrd.  We put it
	   at the highest possible page-aligned address. */
	size = (initrd_size + PAGE - 1) & ~(PAGE - 1);
	for (frag = &frag_list[1]; frag->size != 0; frag++);
	frag--;
	while ((size > 0) && (frag > frag_list)) {
		int step = min(size, frag->size);
		initrd_start = frag->start + frag->size - step;
		size -= step;
		frag--;
	}
	/* XXX miket: trying this out */
#ifndef MSW
	printf("initrd_start is %lx\n", initrd_start);
#else
	sprintf(str, "initrd_start is %lx\n", initrd_start);
	consoleout (str);
#endif
	initrd_start = INITRD_START;
#ifndef MSW
	printf("Moving initrd_start to %lx\n", initrd_start);
#else
	sprintf(str, "Moving initrd_start to %lx\n", initrd_start);
	consoleout (str);
#endif
	/* end XXX area */
	if (initrd_start < kernel_end) {
#ifndef MSW
		printf("Not enough space for initrd\n");
#else
		messagebox("Not enough space for initrd\n");
#endif
		exit(1);
	}

#ifndef MSW
	printf("Loading %s:\n", initrd);
	print_size(initrd_start, initrd_size);
	target_write_fragmented(initrd_start, initrd_buf, initrd_size);
	free(initrd_buf);
#else
	sprintf(str, "Loading %s:\n", initrd);
	consoleout (str);
#endif
	
#ifndef MSW
	printf("Writing parameter area\n");
#else
	consoleout("Writing parameter area\n");
#endif
	target_write_params(initrd_start, initrd_size);
	
	switch(hardware) {
	case 'e':
		post_armadillo9();
		break;
	default:
#ifndef MSW
		fprintf(stderr, "%s: Internal error - invalid hardware value\n",
				progname);
#else
		sprintf(str, "%s: Internal error - invalid hardware value\n",
				progname);
		messagebox (str);
#endif
		exit(1);
	}
	ping();
	
	if (boot) {
		printf("Starting kernel\n");
		put_char('c');
		put_word(DRAM_START + KERNEL_OFFSET);
		put_word(0);	/* sanity check */
		put_word(arch_number);
		put_word(0);
		put_word(0);
	}

#ifndef MSW
	/* I've found this bit useful for debugging: If the kernel doesn't
	   seem to boot, run "shoehorn --terminal" and plug head-armv.S
	   and init/main.c full with things like "IO_UARTDR1 = '!';" or
	   similar.  Then you'll find out how far the setup code gets. */
	if (terminal)
		serial_terminal();
#endif

	serial_close();
	return 0;
}
